/**
 * 
 * Program with simple HTTP server for only necessary processing of GET REST requests
 * 
 * Complete controlling of one axis movement over calling of steps* and free* programs
 * One program for one axis, one axis routed to one socket (IP:port)
 * 
 * Copyright 2019 Luděk Horáček <12@12in.cz>
 * 
 * License GPL
 *
 */

#include <string.h>
#include <stdio.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <unistd.h> 
#include <netinet/in.h> 
#include <time.h> 
#include <sys/wait.h>

#ifndef SELAXIS
#define SELAXIS "axis1.h"
#endif

#include SELAXIS
#define RESPOK  "HTTP/1.1 200 OK\r\nContent-Type: text/plain; charset=utf-8\r\nAccess-Control-Allow-Origin: *\r\nConnection: close\r\n\r\nOK"
#define RESPPOS  "HTTP/1.1 200 OK\r\nContent-Type: text/plain; charset=utf-8\r\nAccess-Control-Allow-Origin: *\r\nConnection: close\r\n\r\n"
#define RESPNO  "HTTP/1.1 200 OK\r\nContent-Type: text/plain; charset=utf-8\r\nAccess-Control-Allow-Origin: *\r\nConnection: close\r\n\r\nNO"
#define RESP404 "HTTP/1.1 204 Not Found\r\nContent-Type: text/plain; charset=utf-8\r\nAccess-Control-Allow-Origin: *\r\nConnection: close"

pid_t  pid;
FILE *fptr;
char *ptr;

void wchild(int sig) {
 wait(NULL);
}

int main(int argc, char **argv) {
 signal(SIGCHLD, wchild);

 unlink(STEPSID);
 unlink(FREEID);

 int listenfd = 0, connfd = 0, valread;
 struct sockaddr_in serv_addr;
 char sendBuff[1024];

 //listenfd = socket(AF_INET, SOCK_STREAM, 0);
 if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) == 0)
 {
  perror("Creating socket failed.");
  exit(EXIT_FAILURE);
 }
 
 
 memset(&serv_addr, '0', sizeof(serv_addr));
 memset(sendBuff, '0', sizeof(sendBuff));

 serv_addr.sin_family = AF_INET;
 serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
 serv_addr.sin_port = htons(PORT);

 //bind(listenfd, (struct sockaddr*)&serv_addr, sizeof(serv_addr));
 if (bind(listenfd, (struct sockaddr*)&serv_addr, sizeof(serv_addr)) < 0)
 {
  perror("Binding failed.");
  exit(EXIT_FAILURE);
 }

 listen(listenfd, 10);

 while(1) {
   char buffer[1024] = {0};
   char requested[256] = {0};
   char response[256] = {0};
   char *poshttp;
   char reqargs[4][32] = {0};
   int i = 0;

   connfd = accept(listenfd, (struct sockaddr*)NULL, NULL);

   valread = read(connfd, buffer, 1024);
   poshttp = strstr(buffer, " HTTP");
   strncpy(requested, buffer, poshttp - buffer);
   char *parts = strtok(requested, "/");
   while (parts != NULL) {
    if (0 == strcmp("GET ", parts)) {
      parts = strtok(NULL, "/");
      continue;
    }
    strcpy(reqargs[i], parts);
    i++;
    parts = strtok(NULL, "/");
   }

   if (0 == strcmp("forward", reqargs[0])) { // Stepping forward
    if (access(STEPSID, F_OK ) != -1) { // only one stepping process is possible
     strcpy(response, RESPNO);
    }
    else if (access(FREEID, F_OK ) != -1) {
     strcpy(response, RESPNO);
    }
    else {
     strcpy(response, RESPOK);
     snprintf(sendBuff, sizeof(sendBuff), "%s\r\n", response);
     write(connfd, sendBuff, strlen(sendBuff));
     close(connfd);

     pid = fork();
     if (0 == pid) {
      execlp(STEPSPROG, SUDOSTEPSPROG, reqargs[1], "0", reqargs[2], reqargs[3], NULL);
      perror("Running forward failed.");
      exit(127);
     }
     else {
      continue;
     }
    }
   }
   else if (0 == strcmp("backward", reqargs[0])) { // Stepping backward
    if (access(STEPSID, F_OK ) != -1) { // only one stepping process is possible
     strcpy(response, RESPNO);
    }
    else if (access(FREEID, F_OK ) != -1) {
     strcpy(response, RESPNO);
    }
    else {
     strcpy(response, RESPOK);
     snprintf(sendBuff, sizeof(sendBuff), "%s\r\n", response);
     write(connfd, sendBuff, strlen(sendBuff));
     close(connfd);

     pid = fork();
     if (0 == pid) {
      execlp(STEPSPROG, SUDOSTEPSPROG, reqargs[1], "1", reqargs[2], reqargs[3], NULL);
      perror("Running backward failed.");
      exit(127);
     }
     else {
      continue;
     }
    }
   }
   else if (0 == strcmp("laxforward", reqargs[0])) { // Free steps forward
    if (access(STEPSID, F_OK ) != -1) { // only one stepping process is possible
     strcpy(response, RESPNO);
    }
    else if (access(FREEID, F_OK ) != -1) {
     strcpy(response, RESPNO);
    }
    else {
     strcpy(response, RESPOK);
     snprintf(sendBuff, sizeof(sendBuff), "%s\r\n", response);
     write(connfd, sendBuff, strlen(sendBuff));
     close(connfd);

     pid = fork();
     if (0 == pid) {
      execlp(LAXPROG, SUDOLAXPROG, reqargs[1], "0", reqargs[2], reqargs[3], NULL);
      perror("Running lax forward failed.");
      exit(127);
     }
     else {
      continue;
     }
    }
   }
   else if (0 == strcmp("laxbackward", reqargs[0])) { // Free steps backward
    if (access(STEPSID, F_OK ) != -1) { // only one stepping process is possible
     strcpy(response, RESPNO);
    }
    else if (access(FREEID, F_OK ) != -1) {
     strcpy(response, RESPNO);
    }
    else {
     strcpy(response, RESPOK);
     snprintf(sendBuff, sizeof(sendBuff), "%s\r\n", response);
     write(connfd, sendBuff, strlen(sendBuff));
     close(connfd);

     pid = fork();
     if (0 == pid) {
      execlp(LAXPROG, SUDOLAXPROG, reqargs[1], "1", reqargs[2], reqargs[3], NULL);
      perror("Running lax backward failed.");
      exit(127);
     }
     else {
      continue;
     }
    }
   }
   else if (0 == strcmp("pos", reqargs[0])) { // get relative position
    char cfc[256], cfch[256];
    unlink(POSACT);

    if (access(STEPSID, F_OK ) != -1) { // get from fresh temporary file
     fptr = fopen(STEPSID, "r");
     fgets(cfc, 256, fptr);
     fclose(fptr);

     pid = strtol(cfc, &ptr, 10);
     kill(pid, SIGUSR1);

     usleep(250);
     while (access(POSACT, R_OK) != 0) usleep(50);

     fptr = fopen(POSACT, "r");
     fgets(cfch, 256, fptr);
     fclose(fptr);
    }
    else { // get from regular file
     fptr = fopen(POSTXT, "r");
     fgets(cfch, 256, fptr);
     fclose(fptr);
    }

    snprintf(response, sizeof(response), "%s%s", RESPPOS, cfch);
   }
   else if (0 == strcmp("zero", reqargs[0])) { // set relative position to zero
    char cfch[1] = "0";
    fptr = fopen(POSTXT, "w");
    fwrite(&cfch, sizeof(cfch), 1, fptr); 
    fclose(fptr);

    strcpy(response, RESPOK);
   }
   else if (0 == strcmp("stop", reqargs[0])) { // stop stepping
    if (access(STEPSID, F_OK ) != -1) { // send SIGTERM to existing stepping process
     char cfch[256];
     fptr = fopen(STEPSID, "r");
     fgets(cfch, 256, fptr);
     fclose(fptr);

     pid = strtol(cfch, &ptr, 10);
     kill(pid, SIGTERM);
    }
    if (access(FREEID, F_OK ) != -1) { // send SIGTERM to existing free steps process
     char cfch[256];
     fptr = fopen(FREEID, "r");
     fgets(cfch, 256, fptr);
     fclose(fptr);

     pid = strtol(cfch, &ptr, 10);
     kill(pid, SIGTERM);
    }

    strcpy(response, RESPOK);
   }
   else if (0 == strcmp("shutdown", reqargs[0])) { // shutdown Raspberry Pi
     system("sudo shutdown -h now"); // shutdown

     strcpy(response, RESPOK);
   }
   else {
     strcpy(response, RESP404);
   }

   snprintf(sendBuff, sizeof(sendBuff), "%s\r\n", response);
   write(connfd, sendBuff, strlen(sendBuff));
   close(connfd);
 }
 return 0;
}
